from typing import AnyStr, List
from pydantic import BaseSettings, Extra, Field

class Time(BaseSettings):
    hour: int = Field(0, alias="HOUR")
    minute: int = Field(0, alias="MINUTE")

    class Config:
        extra = "allow"
        case_sensitive = False
        anystr_lower = True


class Config(BaseSettings):
    # plugin custom config
    plugin_setting: str = "default"

    todo_users: List[str] = []
    todo_groups: List[str] = []

    todo_send_time: List[Time] = []

    class Config:
        extra = Extra.allow
        case_sensitive = False



from typing import AnyStr, List
import jionlp as jio
from typing import Union
import re
import time


def get_time_from_text(text: str) -> (Union[str, None], bool, str):
    try:
        res: dict = jio.parse_time(text,time.time())
        if res['type'] == 'time_point':
            time_point: str = res['time'][0][0:10]
            return time_point, True, ""
        else:
            return None, False, "由于时间非特定时间点，拒绝。"
    except ValueError or RuntimeError:
        return None, False, "没有看到时间，拒绝。"


def get_name_from_text(text: str) -> (Union[str, None], bool, str):
    try:
        if "\"" in text:
            name = "".join(re.findall(r'[\"](.*?)[\"]', text))
        else:
            keyphrases = jio.keyphrase.extract_keyphrase(text)
            if len(keyphrases) > 1:
                name = "".join(keyphrases)
            elif len(keyphrases) == 0:
                return None, False, "未发现事件名，拒绝。可以通过英文双引号来突出事件！"
            else:
                name = keyphrases[0]
        return name, True, ""
    except ValueError or RuntimeError:
        return None, False, "未发现事件名，拒绝。可以通过英文双引号来突出事件！"


priority_5_list = ["重要", "尽快", "一定"]


def get_priority_from_text(text: str) -> (int, bool, str):
    if any(word in text for word in priority_5_list):
        priority = 5
    elif re.search("优先级为", text):
        exact_match = re.search("优先级为[1-5]", text)
        if not exact_match:
            return -1, False, "优先级应当在1-5之间，拒绝"
        else:
            priority = int(exact_match[0][4])
    else:
        priority = 3
    return priority, True, ""



from datetime import datetime
from typing import Union, AnyStr, List
import nonebot
from nonebot import on_keyword
from nonebot import require
from nonebot.adapters.onebot.v11 import MessageSegment, PrivateMessageEvent, GroupMessageEvent
from nonebot.adapters.onebot.v11.permission import GROUP_ADMIN, GROUP_OWNER, PRIVATE_FRIEND
import time

from .nlp_util import get_time_from_text, get_name_from_text, get_priority_from_text
from .todo import TodoToken, TodoUtil
from .config import Config

global_config = nonebot.get_driver().config
plugin_config = Config(**global_config.dict())

add_todo = on_keyword({'提醒', 'nonebot_todo'}, permission=GROUP_ADMIN | GROUP_OWNER | PRIVATE_FRIEND)
remove_todo = on_keyword({'删除', '去掉'}, permission=GROUP_ADMIN | GROUP_OWNER | PRIVATE_FRIEND)
finish_todo = on_keyword({'完成'}, permission=GROUP_ADMIN | GROUP_OWNER | PRIVATE_FRIEND)
change_todo_time = on_keyword({'更正', '改'}, permission=GROUP_ADMIN | GROUP_OWNER | PRIVATE_FRIEND)
get_todo_pic = on_keyword({'获取todo'}, permission=GROUP_ADMIN | GROUP_OWNER | PRIVATE_FRIEND)
scheduler = require("nonebot_plugin_apscheduler").scheduler


@add_todo.handle()
async def add_todo_handle(event: Union[PrivateMessageEvent, GroupMessageEvent]):
    text: str = event.raw_message
    exceed_time, time_valid, time_error_doc = get_time_from_text(text)
    name, name_valid, name_error_doc = get_name_from_text(text)
    priority, priority_valid, priority_error_doc = get_priority_from_text(text)
    if not time_valid:
        await add_todo.finish(time_error_doc)
    elif not name_valid:
        await add_todo.finish(name_error_doc)
    elif not priority_valid:
        await add_todo.finish(priority_error_doc)
    if isinstance(event, PrivateMessageEvent):
        todo_util = TodoUtil(str(event.user_id))
    else:
        todo_util = TodoUtil(str(event.group_id))
    todo_token = TodoToken(name, datetime.today(), exceed_time, priority)
    if todo_util.add_data_to_list(todo_token):
        await add_todo.send(f"已将{name}加入清单，ddl为{exceed_time}，优先级为{priority}。\n当前共{todo_util.list_size}项待办。")
        img = await todo_util.get_list_img()
        await remove_todo.finish(MessageSegment.image(img))
    else:
        await add_todo.finish(f"{name}已在清单中。")


@remove_todo.handle()
async def remove_todo_handle(event: Union[PrivateMessageEvent, GroupMessageEvent]):
    text: str = event.raw_message
    name, name_valid, name_error_doc = get_name_from_text(text)
    if isinstance(event, PrivateMessageEvent):
        todo_util = TodoUtil(str(event.user_id))
    else:
        todo_util = TodoUtil(str(event.group_id))
    if not name_valid:
        await remove_todo.finish(name_error_doc)
    elif todo_util.remove_data(name) == 0:
        await remove_todo.finish(f"不存在名为{name}的表项。")
    else:
        img = await todo_util.get_list_img()
        await remove_todo.send(f"已删除{name}。")
        await remove_todo.finish(MessageSegment.image(img))


@finish_todo.handle()
async def finish_todo_handle(event: Union[PrivateMessageEvent, GroupMessageEvent]):
    text: str = event.raw_message
    name, name_valid, name_error_doc = get_name_from_text(text)
    if isinstance(event, PrivateMessageEvent):
        todo_util = TodoUtil(str(event.user_id))
    else:
        todo_util = TodoUtil(str(event.group_id))
    if not name_valid:
        await remove_todo.finish(name_error_doc)
    elif todo_util.finish_job(name) == 0:
        await remove_todo.finish(f"不存在名为{name}的表项。")
    else:
        await remove_todo.send(f"已完成{name}。还剩{todo_util.list_size}项。")
        img = await todo_util.get_list_img()
        await remove_todo.finish(MessageSegment.image(img))


@change_todo_time.handle()
async def change_todo_time_handle(event: Union[PrivateMessageEvent, GroupMessageEvent]):
    text: str = event.raw_message
    name, name_valid, name_error_doc = get_name_from_text(text)
    expire_time, time_valid, time_error_doc = get_time_from_text(text)
    if isinstance(event, PrivateMessageEvent):
        todo_util = TodoUtil(str(event.user_id))
    else:
        todo_util = TodoUtil(str(event.group_id))
    if not time_valid:
        await add_todo.finish(time_error_doc)
    elif not name_valid:
        await add_todo.finish(name_error_doc)
    else:
        changed_count = todo_util.change_time(name, expire_time)
        if not changed_count == 0:
            await change_todo_time.send(f"将{name}改至{expire_time}")
            img = await todo_util.get_list_img()
            await change_todo_time.finish(MessageSegment.image(img))
        else:
            await change_todo_time.finish(f"没有名为{name}的项目。")


@get_todo_pic.handle()
async def _handle(event: Union[PrivateMessageEvent, GroupMessageEvent]):
    if isinstance(event, PrivateMessageEvent):
        todo_util = TodoUtil(str(event.user_id))
    else:
        todo_util = TodoUtil(str(event.group_id))
    img = await todo_util.get_list_img()
    await get_todo_pic.finish(MessageSegment.image(img))


async def send_todo():
    for user in plugin_config.todo_users:
        todo_util = TodoUtil(str(user))
        img = await todo_util.get_list_img()
        await nonebot.get_bot().send_private_msg(user_id=user, message=MessageSegment.image(img))
        time.sleep(5)
        # 防止风控

    for group in plugin_config.todo_groups:
        todo_util = TodoUtil(str(group))
        img = await todo_util.get_list_img()
        await nonebot.get_bot().send_group_msg(group_id=group, message=MessageSegment.image(img))
        time.sleep(5)
        # 防止风控


for index, _time in enumerate(plugin_config.todo_send_time):
    scheduler.add_job(send_todo, "cron", hour=_time.hour, minute=_time.minute, id=str(f"send_todo_for_{index}"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_todo_nlp']

package_data = \
{'': ['*'], 'nonebot_plugin_todo_nlp': ['templates/*', 'templates/css/*']}

install_requires = \
['jionlp>=1.4.17,<2.0.0',
 'nonebot-adapter-onebot>=2.1.3,<3.0.0',
 'nonebot-plugin-apscheduler==0.2.0',
 'nonebot-plugin-htmlrender>=0.1.1,<0.2.0',
 'nonebot2>=2.0.0-beta.5,<3.0.0',
 'pandas>=1.4.3,<2.0.0',
 'pathlib>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-todo-nlp',
    'version': '0.1.8rc0',
    'description': '一款自动识别提醒内容，可生成todo图片并定时推送的nonebot2插件',
    'long_description': 'None',
    'author': 'CofinCup',
    'author_email': '864341840@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import AnyStr, List
import pandas as pd
import pandas.errors
import os
from nonebot_plugin_htmlrender import template_to_pic
from pandas import DataFrame, read_csv, concat
from datetime import datetime, timedelta
from typing import Union
from re import match
from pathlib import Path
from pandas.errors import EmptyDataError


# TODO:add log

class TodoToken:
    def __init__(self,
                 name: str,
                 start_date: Union[str, datetime, None] = None,
                 end_date: Union[str, datetime, None] = None,
                 priority: Union[int, None] = 3,
                 format_str: str = "%Y-%m-%d"):
        self.format_str = format_str
        self._name = name
        self._start_date = TodoToken.time_parse(start_date, format_str)
        self._end_date = TodoToken.time_parse(end_date, format_str)
        self._priority = priority

    @property
    def name(self):
        return self._name

    @property
    def start_date(self):
        return self._start_date

    @property
    def end_date(self):
        return self._end_date

    @property
    def priority(self):
        return self._priority

    @property
    def time_left(self):
        end = datetime.strptime(self.end_date, self.format_str) + timedelta(days=1)
        today = datetime.today()
        time_len = (end - today).days
        return time_len

    @property
    def percentage(self):
        start = datetime.strptime(self.start_date, self.format_str)
        end = datetime.strptime(self.end_date, self.format_str) + timedelta(days=1)
        today = datetime.today()
        time_len = (end - start)
        time_spent = (today - start)
        return int(time_spent / time_len * 100)

    @classmethod
    def time_parse(cls, _time: [str, datetime, None],
                   format_str: str = "%Y-%m-%d"):
        if _time is None:
            _time = datetime.today()
        elif isinstance(_time, datetime):
            _time = _time.strftime(format_str)
        else:
            try:
                test = datetime.strptime(_time, format_str)
            except ValueError:
                raise ValueError("Time parsing failed.")
        return _time

    def __lt__(self, other):
        if not isinstance(other, TodoToken):
            raise ValueError("{other} is not a TodoToken.")
        if self.end_date != other.end_date:
            return self.end_date < other.end_date
        elif self.priority != other.priority:
            return self.priority < other.priority
        elif self.start_date != other.start_date:
            return self.start_date < other.start_date
        else:
            return self.name < other.name

    def __eq__(self, other):
        if not isinstance(other, TodoToken):
            raise ValueError("{other} is not a TodoToken.")
        if (self.name == other.name and
                self.start_date == other.start_date and
                self.end_date == other.end_date and
                self.priority == other.priority):
            return True
        else:
            return False

    def __str__(self):
        return self._name + ',' + self._start_date + ',' + self._end_date + ',' + str(self._priority)

    def to_dict(self):
        return {'name': self._name,
                'start_date': self._start_date,
                'end_date': self._end_date,
                'priority': self._priority}


class TodoList:
    def __init__(self, user: str, path: str, format_str: str = "%Y-%m-%d"):
        self._list = list()
        self._user = user
        self._path = path
        self._format_str = format_str
        try:
            df = read_csv(self.path + '/' + user + ".csv")
        except FileNotFoundError or EmptyDataError:
            df = None
        if df is not None:
            for index, row in df.iterrows():
                token = TodoToken(row['name'], row['start_date'],
                                  row['end_date'], row['priority'], self.format_str)
                self._list.append(token)

    @property
    def size(self):
        return len(self._list)

    @property
    def path(self):
        return self._path

    @property
    def format_str(self):
        return self._format_str

    def add_data(self, token: TodoToken) -> bool:
        if token in self._list:
            return False
        else:
            self._list.append(token)
            self.write_data()
            return True

    def remove_data(self, name: str) -> int:
        list_len = len(self._list)
        self._list = [token for token in self._list if not match(name, token.name)]
        if list_len > len(self._list):
            self.write_data()
        return list_len - len(self._list)

    def change_data(self, name: str, slot: str, content: str):
        count = 0
        if slot == 'priority':
            for token in self._list:
                if match(name, token.name) is not None:
                    token._priority = int(content)
                    count += 1
        else:
            for token in self._list:
                if match(name, token.name) is not None:
                    setattr(token, slot, content)
                    count += 1
        self.write_data()
        return count

    def write_data(self):
        self._list.sort()
        if len(self._list) == 0:
            df = pd.DataFrame(columns=['name', 'start_date', 'end_date', 'priority'])
            df.to_csv(self.path + '/' + self._user + '.csv', index=False)

        else:
            df = DataFrame()
            for token in self._list:
                tmp = DataFrame([token.to_dict()])
                df = concat([df, tmp], axis=0, ignore_index=True)
            df.to_csv(self.path + '/' + self._user + '.csv', index=False)

    async def get_img(self):
        list_by_date: List[(str, int, List[TodoToken])] = list()
        now_date = ""
        date_count = 0
        now_date_token: List[TodoToken] = list()
        for token in self._list:
            if token.end_date != now_date:
                date_count += 1
                now_date = token.end_date
                now_date_token:List[TodoToken] = list()
                list_by_date.append((now_date, token.time_left, now_date_token))
            now_date_token.append(token)

        content_width = 500
        content_height = 76.4 + \
                         34 * date_count + \
                         64 * len(self._list)
        # estimated title_height + date_height + list_token_height
        img_width = content_width / 0.618
        img_height = content_height / 0.618
        template_path = str(Path(__file__).parent / "templates")
        img = await template_to_pic(template_path=template_path,
                                    template_name="template.html",
                                    templates={
                                        "list_by_date": list_by_date,
                                    },
                                    pages={
                                        "viewport": {"width": int(img_width), "height": int(img_height)},
                                        "base_url": f"file://{template_path}"
                                    },
                                    )
        return img


class TodoUtil:
    def __init__(self, user: str, path: str = str(Path(__file__).parent / "todo_data" ), format_str: str = "%Y-%m-%d"):
        if not os.path.exists(path):
            os.makedirs(path)
        self._todo_list: TodoList = TodoList(user, path)
        self._format_str = format_str

    @property
    def format_str(self):
        return self._format_str

    @property
    def list_size(self):
        return self.todo_list.size

    @property
    def todo_list(self):
        return self._todo_list

    def add_data_to_list(self, token: TodoToken):
        return self.todo_list.add_data(token)

    def remove_data(self, name: str):
        return self.todo_list.remove_data(name)

    def change_data(self, name: str, slot: str, content: str):
        return self.todo_list.change_data(name, slot, content)

    def change_time(self, name: str, time: str):
        return self.todo_list.change_data(name, '_end_date', time)

    def finish_job(self, name: str):
        return self.remove_data(name)

    def get_list(self):
        return self.todo_list

    def get_list_img(self):
        return self.todo_list.get_img()
